use cgmath::Point2;
use content::Ship;
use object::{ship::ShipPersonality, GameData};
use std::time::Instant;
use winit::event::{ElementState, MouseButton, MouseScrollDelta, TouchPhase, VirtualKeyCode};

use crate::camera::Camera;
use crate::inputstatus::InputStatus;

use galactica_behavior::{behavior, ShipBehavior};
use galactica_constants;
use galactica_content as content;
use galactica_gameobject as object;
use galactica_render::RenderState;
use galactica_world::{util, ParticleBuilder, ShipPhysicsHandle, World};

pub struct Game {
	input: InputStatus,
	last_update: Instant,
	player: ShipPhysicsHandle,
	paused: bool,
	time_scale: f32,
	start_instant: Instant,
	camera: Camera,

	// TODO: include system in world
	//system: object::System,
	shipbehaviors: Vec<Box<dyn ShipBehavior>>,
	playerbehavior: behavior::Player,

	gamedata: GameData,

	content: content::Content,
	world: World,
	new_particles: Vec<ParticleBuilder>,
}

impl Game {
	pub fn new(ct: content::Content) -> Self {
		let mut physics = World::new();
		let mut gamedata = GameData::new();

		let ss = ct.get_ship(content::ShipHandle { index: 0 });

		let mut o1 = object::OutfitSet::new(ss);
		o1.add(&ct, content::OutfitHandle { index: 0 });
		o1.add(&ct, content::OutfitHandle { index: 1 });
		o1.add_gun(&ct, content::GunHandle { index: 0 });
		o1.add_gun(&ct, content::GunHandle { index: 0 });
		o1.add_gun(&ct, content::GunHandle { index: 0 });

		let mut o1 = object::OutfitSet::new(ss);
		o1.add(&ct, content::OutfitHandle { index: 0 });
		o1.add_gun(&ct, content::GunHandle { index: 0 });

		gamedata.create_ship(
			&ct,
			content::ShipHandle { index: 0 },
			content::FactionHandle { index: 0 },
			ShipPersonality::Player,
			o1,
		);

		gamedata.create_ship(
			&ct,
			content::ShipHandle { index: 0 },
			content::FactionHandle { index: 1 },
			ShipPersonality::Dummy,
			object::OutfitSet::new(ss),
		);

		gamedata.create_ship(
			&ct,
			content::ShipHandle { index: 0 },
			content::FactionHandle { index: 1 },
			ShipPersonality::Point,
			o1,
		);

		//let mut shipbehaviors: Vec<Box<dyn ShipBehavior>> = Vec::new();
		//shipbehaviors.push(Box::new(behavior::Dummy::new(h2)));
		//shipbehaviors.push(Box::new(behavior::Point::new(h3)));

		Game {
			last_update: Instant::now(),
			input: InputStatus::new(),
			player: h1,
			start_instant: Instant::now(),

			camera: Camera {
				pos: (0.0, 0.0).into(),
				zoom: 500.0,
				aspect: 1.0,
			},
			//system: object::System::new(&ct, SystemHandle { index: 0 }),
			paused: false,
			time_scale: 1.0,
			world: physics,
			shipbehaviors,
			gamedata,
			content: ct,
			playerbehavior: behavior::Player::new(h1),
			new_particles: Vec::new(),
		}
	}

	pub fn set_camera_aspect(&mut self, v: f32) {
		self.camera.aspect = v
	}

	pub fn process_key(&mut self, state: &ElementState, key: &VirtualKeyCode) {
		self.input.process_key(state, key)
	}

	pub fn process_click(&mut self, state: &ElementState, key: &MouseButton) {
		self.input.process_click(state, key)
	}

	pub fn process_scroll(&mut self, delta: &MouseScrollDelta, phase: &TouchPhase) {
		self.input.process_scroll(delta, phase)
	}

	pub fn set_paused(&mut self, pause: bool) {
		if pause {
			self.paused = true;
			self.input.release_all()
		} else {
			self.paused = false;
		}
	}

	pub fn update(&mut self) {
		let t: f32 = self.last_update.elapsed().as_secs_f32() * self.time_scale;

		self.playerbehavior.key_guns = self.input.key_guns;
		self.playerbehavior.key_thrust = self.input.key_thrust;
		self.playerbehavior.key_right = self.input.key_right;
		self.playerbehavior.key_left = self.input.key_left;
		self.playerbehavior
			.update_controls(&mut self.world, &self.content);

		self.shipbehaviors.retain_mut(|b| {
			// Remove shipbehaviors of destroyed ships
			if self.world.get_ship_mut(&b.get_handle()).is_none() {
				false
			} else {
				b.update_controls(&mut self.world, &self.content);
				true
			}
		});

		self.world.step(t, &self.content, &mut self.new_particles);

		if self.input.v_scroll != 0.0 {
			self.camera.zoom = (self.camera.zoom + self.input.v_scroll)
				.clamp(galactica_constants::ZOOM_MIN, galactica_constants::ZOOM_MAX);
			self.input.v_scroll = 0.0;
		}

		let r = self
			.world
			.get_ship_mut(&self.player)
			.unwrap()
			.physics_handle;
		let r = self.world.get_rigid_body(r.0).unwrap(); // TODO: r.0 shouldn't be public
		let ship_pos = util::rigidbody_position(r);
		self.camera.pos = ship_pos;
		self.last_update = Instant::now();
	}

	pub fn get_frame_state(&mut self) -> RenderState {
		RenderState {
			camera_pos: self.camera.pos,
			camera_zoom: self.camera.zoom,
			current_time: self.start_instant.elapsed().as_secs_f32(),
			content: &self.content,
			world: &self.world,
			particles: &mut self.new_particles,
			player: &self.player,
		}
	}
}
