mod camera;
mod game;
mod inputstatus;

pub use galactica_content as content;

use anyhow::{bail, Result};
use galactica_constants::{self, ASSET_CACHE};
use std::{
	fs,
	path::{Path, PathBuf},
};
use winit::{
	event::{Event, KeyboardInput, WindowEvent},
	event_loop::{ControlFlow, EventLoop},
	window::WindowBuilder,
};

fn main() -> Result<()> {
	let cache_dir = Path::new(ASSET_CACHE);
	fs::create_dir_all(cache_dir)?;

	let atlas_index = cache_dir.join("spriteatlas.toml");

	if !atlas_index.exists() {
		bail!("Could not find sprite atlas!");
	}

	// TODO: pretty error if missing
	let content = content::Content::load_dir(
		PathBuf::from(galactica_constants::CONTENT_ROOT),
		PathBuf::from(galactica_constants::IMAGE_ROOT),
		atlas_index,
		galactica_constants::STARFIELD_SPRITE_NAME.to_owned(),
	)?;

	let event_loop = EventLoop::new();
	let window = WindowBuilder::new().build(&event_loop).unwrap();
	let mut gpu = pollster::block_on(galactica_render::GPUState::new(window, &content))?;
	gpu.init();

	let mut game = game::Game::new(content);
	gpu.update_starfield_buffer();
	game.set_camera_aspect(gpu.window_size.width as f32 / gpu.window_size.height as f32);

	event_loop.run(move |event, _, control_flow| {
		match event {
			Event::RedrawRequested(window_id) if window_id == gpu.window().id() => {
				match gpu.render(game.get_frame_state()) {
					Ok(_) => {}
					Err(wgpu::SurfaceError::Lost) => gpu.resize(),
					Err(wgpu::SurfaceError::OutOfMemory) => *control_flow = ControlFlow::Exit,
					// All other errors (Outdated, Timeout) should be resolved by the next frame
					Err(e) => eprintln!("{:?}", e),
				}
			}

			Event::MainEventsCleared => {
				game.update();
				gpu.window().request_redraw();
			}

			Event::WindowEvent {
				ref event,
				window_id,
			} if window_id == gpu.window.id() => match event {
				WindowEvent::Focused(state) => {
					game.set_paused(!state);
				}
				WindowEvent::CloseRequested => *control_flow = ControlFlow::Exit,
				WindowEvent::KeyboardInput {
					input:
						KeyboardInput {
							state,
							virtual_keycode: Some(key),
							..
						},
					..
				} => game.process_key(state, key),
				WindowEvent::MouseInput { state, button, .. } => {
					game.process_click(state, button);
				}
				WindowEvent::MouseWheel { delta, phase, .. } => {
					game.process_scroll(delta, phase);
				}
				WindowEvent::Resized(_) => {
					gpu.resize();
					game.set_camera_aspect(
						gpu.window_size.width as f32 / gpu.window_size.height as f32,
					);
				}
				WindowEvent::ScaleFactorChanged { .. } => {
					gpu.resize();
					game.set_camera_aspect(
						gpu.window_size.width as f32 / gpu.window_size.height as f32,
					);
				}
				_ => {}
			},
			_ => {}
		}
	});
}
