//! GPUState routines for drawing HUD elements

use std::f32::consts::TAU;

use cgmath::{Deg, InnerSpace, Point2, Rad, Vector2};
use galactica_world::util;

use crate::{
	vertexbuffer::{
		types::{RadialBarInstance, UiInstance},
		BufferObject,
	},
	GPUState, PositionAnchor, RenderState,
};

impl GPUState {
	pub(super) fn hud_add_radar(&mut self, state: &RenderState) {
		let radar_range = 4000.0;
		let radar_size = 300.0;
		let hide_range = 0.85;
		let shrink_distance = 20.0;
		//let system_object_scale = 1.0 / 600.0;
		let ship_scale = 1.0 / 10.0;

		let (_, player_body) = state.world.get_ship_body(*state.player).unwrap();
		let player_position = util::rigidbody_position(player_body);
		//let planet_sprite = state.content.get_sprite_handle("ui::planetblip");
		let ship_sprite = state.content.get_sprite_handle("ui::shipblip");
		let arrow_sprite = state.content.get_sprite_handle("ui::centerarrow");

		// Enforce buffer limit
		if self.vertex_buffers.ui_counter as u64 > galactica_constants::UI_SPRITE_INSTANCE_LIMIT {
			// TODO: no panic, handle this better.
			panic!("UI limit exceeded!")
		}

		// Push this object's instance
		self.queue.write_buffer(
			&self.vertex_buffers.ui.instances,
			UiInstance::SIZE * self.vertex_buffers.ui_counter,
			bytemuck::cast_slice(&[UiInstance {
				anchor: PositionAnchor::NwNw.to_int(),
				position: [10.0, -10.0],
				angle: 0.0,
				size: radar_size,
				color: [1.0, 1.0, 1.0, 1.0],
				sprite_index: state.content.get_sprite_handle("ui::radar").get_index(),
			}]),
		);
		self.vertex_buffers.ui_counter += 1;

		/*
		// Draw system objects
		for o in &system.bodies {
			let size = (o.size / o.pos.z) / (radar_range * system_object_scale);
			let p = Point2 {
				x: o.pos.x,
				y: o.pos.y,
			};
			let d = (p - player_position) / radar_range;
			// Add half the blip sprite's height to distance
			let m = d.magnitude() + (size / (2.0 * radar_size));
			if m < hide_range {
				// Shrink blips as they get closeto the edge
				let size = size.min((hide_range - m) * size * shrink_distance);
				if size <= 2.0 {
					// Don't draw super tiny sprites, they flicker
					continue;
				}
				out.push(UiSprite {
					sprite: planet_sprite,
					pos: AnchoredUiPosition::NwC(
						Point2 {
							x: radar_size / 2.0 + 10.0,
							y: radar_size / -2.0 - 10.0,
						} + (d * (radar_size / 2.0)),
					),
					dimensions: Point2 {
						x: planet_sprite.aspect,
						y: 1.0,
					} * size,
					angle: o.angle,
					color: Some([0.5, 0.5, 0.5, 1.0]),
				});
			}
		}
		*/

		// Draw ships
		for (s, r) in state.world.iter_ship_body() {
			let ship = state.content.get_ship(s.ship.handle);
			let size = (ship.size * ship.sprite.aspect) * ship_scale;
			let p = util::rigidbody_position(r);
			let d = (p - player_position) / radar_range;
			let m = d.magnitude() + (size / (2.0 * radar_size));
			if m < hide_range {
				let size = size.min((hide_range - m) * size * shrink_distance);
				if size < 2.0 {
					continue;
				}
				let angle = util::rigidbody_rotation(r).angle(Vector2 { x: 0.0, y: 1.0 });
				let f = state.content.get_faction(s.ship.faction).color;
				let f = [f[0], f[1], f[2], 1.0];

				let position = Point2 {
					x: radar_size / 2.0 + 10.0,
					y: radar_size / -2.0 - 10.0,
				} + (d * (radar_size / 2.0));

				// Enforce buffer limit
				if self.vertex_buffers.ui_counter as u64
					> galactica_constants::UI_SPRITE_INSTANCE_LIMIT
				{
					// TODO: no panic, handle this better.
					panic!("UI limit exceeded!")
				}

				// Push this object's instance
				self.queue.write_buffer(
					&self.vertex_buffers.ui.instances,
					UiInstance::SIZE * self.vertex_buffers.ui_counter,
					bytemuck::cast_slice(&[UiInstance {
						anchor: PositionAnchor::NwC.to_int(),
						position: position.into(),
						angle: -angle.0, // TODO: consistent angles
						size,
						color: f.into(),
						sprite_index: ship_sprite.get_index(),
					}]),
				);
				self.vertex_buffers.ui_counter += 1;
			}
		}

		// Draw viewport frame
		let d = Vector2 {
			x: (state.camera_zoom / 2.0) * self.window_aspect,
			y: state.camera_zoom / 2.0,
		} / radar_range;
		let m = d.magnitude();
		let d = d * (radar_size / 2.0);
		let color = [0.3, 0.3, 0.3, 1.0];
		if m < 0.8 {
			let sprite = state.content.get_sprite_handle("ui::radarframe");
			let size = 7.0f32.min((0.8 - m) * 70.0);

			// Enforce buffer limit (this section adds four items)
			if self.vertex_buffers.ui_counter as u64 + 4
				> galactica_constants::UI_SPRITE_INSTANCE_LIMIT
			{
				// TODO: no panic, handle this better.
				panic!("UI limit exceeded!")
			}

			self.queue.write_buffer(
				&self.vertex_buffers.ui.instances,
				UiInstance::SIZE * self.vertex_buffers.ui_counter,
				bytemuck::cast_slice(&[UiInstance {
					anchor: PositionAnchor::NwNw.to_int(),
					position: Point2 {
						x: (radar_size / 2.0 + 10.0) - d.x,
						y: (radar_size / -2.0 - 10.0) + d.y,
					}
					.into(),
					angle: 0.0,
					size,
					color,
					sprite_index: sprite.get_index(),
				}]),
			);
			self.vertex_buffers.ui_counter += 1;

			self.queue.write_buffer(
				&self.vertex_buffers.ui.instances,
				UiInstance::SIZE * self.vertex_buffers.ui_counter,
				bytemuck::cast_slice(&[UiInstance {
					anchor: PositionAnchor::NwSw.to_int(),
					position: Point2 {
						x: (radar_size / 2.0 + 10.0) - d.x,
						y: (radar_size / -2.0 - 10.0) - d.y,
					}
					.into(),
					angle: Rad::from(Deg(90.0)).0,
					size,
					color,
					sprite_index: sprite.get_index(),
				}]),
			);
			self.vertex_buffers.ui_counter += 1;

			self.queue.write_buffer(
				&self.vertex_buffers.ui.instances,
				UiInstance::SIZE * self.vertex_buffers.ui_counter,
				bytemuck::cast_slice(&[UiInstance {
					anchor: PositionAnchor::NwSe.to_int(),
					position: Point2 {
						x: (radar_size / 2.0 + 10.0) + d.x,
						y: (radar_size / -2.0 - 10.0) - d.y,
					}
					.into(),
					angle: Rad::from(Deg(180.0)).0,
					size,
					color,
					sprite_index: sprite.get_index(),
				}]),
			);
			self.vertex_buffers.ui_counter += 1;

			self.queue.write_buffer(
				&self.vertex_buffers.ui.instances,
				UiInstance::SIZE * self.vertex_buffers.ui_counter,
				bytemuck::cast_slice(&[UiInstance {
					anchor: PositionAnchor::NwNe.to_int(),
					position: Point2 {
						x: (radar_size / 2.0 + 10.0) + d.x,
						y: (radar_size / -2.0 - 10.0) + d.y,
					}
					.into(),
					angle: Rad::from(Deg(270.0)).0,
					size,
					color,
					sprite_index: sprite.get_index(),
				}]),
			);
			self.vertex_buffers.ui_counter += 1;
		}

		// Arrow to center of system
		let q = Point2 { x: 0.0, y: 0.0 } - player_position;
		let m = q.magnitude();
		if m > 200.0 {
			let player_angle = q.angle(Vector2 { x: 0.0, y: 1.0 });

			let position: Point2<f32> = Point2 {
				x: radar_size / 2.0 + 10.0,
				y: radar_size / -2.0 - 10.0,
			} + ((q.normalize() * 0.865) * (radar_size / 2.0));

			if self.vertex_buffers.ui_counter as u64 > galactica_constants::UI_SPRITE_INSTANCE_LIMIT
			{
				// TODO: no panic, handle this better.
				panic!("UI limit exceeded!")
			}

			self.queue.write_buffer(
				&self.vertex_buffers.ui.instances,
				UiInstance::SIZE * self.vertex_buffers.ui_counter,
				bytemuck::cast_slice(&[UiInstance {
					anchor: PositionAnchor::NwC.to_int(),
					position: position.into(),
					angle: -player_angle.0,
					size: 10.0,
					color: [1.0, 1.0, 1.0, 1f32.min((m - 200.0) / 400.0)],
					sprite_index: arrow_sprite.get_index(),
				}]),
			);
			self.vertex_buffers.ui_counter += 1;
		}
	}

	pub(super) fn hud_add_status(&mut self, state: &RenderState) {
		if self.vertex_buffers.ui_counter as u64 > galactica_constants::UI_SPRITE_INSTANCE_LIMIT {
			// TODO: no panic, handle this better.
			panic!("UI limit exceeded!")
		}

		let (s, _) = state.world.get_ship_body(*state.player).unwrap();
		let max_shields = s.ship.outfits.stat_sum().shield_strength;
		let current_shields = s.ship.shields;
		let current_hull = s.ship.hull;
		let max_hull = state.content.get_ship(s.ship.handle).hull;

		self.queue.write_buffer(
			&self.vertex_buffers.ui.instances,
			UiInstance::SIZE * self.vertex_buffers.ui_counter,
			bytemuck::cast_slice(&[UiInstance {
				anchor: PositionAnchor::NeNe.to_int(),
				position: [-10.0, -10.0],
				angle: 0.0,
				size: 200.0,
				color: [1.0, 1.0, 1.0, 1.0],
				sprite_index: state.content.get_sprite_handle("ui::status").get_index(),
			}]),
		);
		self.vertex_buffers.ui_counter += 1;

		// We add two items here, so +2
		if self.vertex_buffers.radialbar_counter as u64 + 2
			> galactica_constants::RADIALBAR_SPRITE_INSTANCE_LIMIT
		{
			// TODO: no panic, handle this better.
			panic!("Radialbar limit exceeded!")
		}

		self.queue.write_buffer(
			&self.vertex_buffers.radialbar.instances,
			RadialBarInstance::SIZE * self.vertex_buffers.radialbar_counter,
			bytemuck::cast_slice(&[RadialBarInstance {
				position: [-19.0, -19.0],
				anchor: PositionAnchor::NeNe.to_int(),
				diameter: 182.0,
				stroke: 5.0,
				color: [0.3, 0.6, 0.8, 1.0],
				angle: (current_shields / max_shields) * TAU,
			}]),
		);
		self.vertex_buffers.radialbar_counter += 1;

		self.queue.write_buffer(
			&self.vertex_buffers.radialbar.instances,
			RadialBarInstance::SIZE * self.vertex_buffers.radialbar_counter,
			bytemuck::cast_slice(&[RadialBarInstance {
				position: [-27.0, -27.0],
				anchor: PositionAnchor::NeNe.to_int(),
				diameter: 166.0,
				stroke: 5.0,
				color: [0.8, 0.7, 0.5, 1.0],
				angle: (current_hull / max_hull) * TAU,
			}]),
		);
		self.vertex_buffers.radialbar_counter += 1;
	}
}
