#![warn(missing_docs)]

//! This module keeps track of the visible world.
//! Ships, projectiles, collisions, etc.

pub mod objects;
pub mod util;
mod world;
mod wrapper;

use cgmath::{Matrix2, Point2, Rad, Vector2};
use galactica_content as content;
use rand::Rng;
pub use world::World;

use rapier2d::{dynamics::RigidBodyHandle, geometry::ColliderHandle};

/// A lightweight handle for a specific ship in our physics system
#[derive(Debug, Copy, Clone, PartialEq, Eq)]
pub struct ShipPhysicsHandle(pub RigidBodyHandle, ColliderHandle);

/// Instructions to create a new particle
pub struct ParticleBuilder {
	/// The sprite to use for this particle
	pub sprite: content::SpriteHandle,

	/// This object's center, in world coordinates.
	pub pos: Point2<f32>,

	/// This particle's velocity, in world coordinates
	pub velocity: Vector2<f32>,

	/// This particle's angle
	pub angle: Rad<f32>,

	/// This particle's angular velocity (rad/sec)
	pub angvel: Rad<f32>,

	/// This particle's lifetime, in seconds
	pub lifetime: f32,

	/// The size of this particle,
	/// given as height in world units.
	pub size: f32,

	/// Fade this particle out over this many seconds as it expires
	pub fade: f32,
}

impl ParticleBuilder {
	/// Create a ParticleBuilder from an Effect
	pub fn from_content(
		effect: &content::Effect,
		pos: Point2<f32>,
		parent_angle: Rad<f32>,
		parent_velocity: Vector2<f32>,
		target_velocity: Vector2<f32>,
	) -> Self {
		let mut rng = rand::thread_rng();

		let velocity = {
			let a =
				rng.gen_range(-effect.velocity_scale_parent_rng..=effect.velocity_scale_parent_rng);
			let b =
				rng.gen_range(-effect.velocity_scale_target_rng..=effect.velocity_scale_target_rng);

			let velocity = ((effect.velocity_scale_parent + a) * parent_velocity)
				+ ((effect.velocity_scale_target + b) * target_velocity);

			Matrix2::from_angle(Rad(
				rng.gen_range(-effect.direction_rng..=effect.direction_rng)
			)) * velocity
		};

		// Rad has odd behavior when its angle is zero, so we need extra checks here
		let angvel = if effect.angvel_rng == 0.0 {
			effect.angvel
		} else {
			Rad(effect.angvel.0 + rng.gen_range(-effect.angvel_rng..=effect.angvel_rng))
		};
		let angle = if effect.angle_rng == 0.0 {
			parent_angle + effect.angle
		} else {
			parent_angle + effect.angle + Rad(rng.gen_range(-effect.angle_rng..=effect.angle_rng))
		};

		ParticleBuilder {
			sprite: effect.sprite,
			pos,
			velocity,

			angle,
			angvel,

			lifetime: 0f32
				.max(effect.lifetime + rng.gen_range(-effect.lifetime_rng..=effect.lifetime_rng)),

			// Make sure size isn't negative. This check should be on EVERY rng!
			size: 0f32.max(effect.size + rng.gen_range(-effect.size_rng..=effect.size_rng)),

			fade: 0f32.max(effect.fade + rng.gen_range(-effect.fade_rng..=effect.fade_rng)),
		}
	}
}
