use cgmath::{Deg, EuclideanSpace, InnerSpace, Matrix2, Point2, Rad, Vector2, Zero};
use crossbeam::channel::Receiver;
use nalgebra::{point, vector};
use rand::Rng;
use rapier2d::{
	dynamics::{RigidBody, RigidBodyBuilder, RigidBodyHandle},
	geometry::{ColliderBuilder, ColliderHandle, CollisionEvent},
	pipeline::{ActiveEvents, ChannelEventCollector},
};
use std::{collections::HashMap, f32::consts::PI};

use crate::{
	objects,
	objects::{ProjectileWorldObject, ShipWorldObject},
	util,
	wrapper::Wrapper,
	ParticleBuilder, ShipPhysicsHandle,
};
use galactica_content as content;
use galactica_gameobject as object;

/// Keeps track of all objects in the world that we can interact with.
/// Also wraps our physics engine
pub struct World {
	wrapper: Wrapper,
	projectiles: HashMap<ColliderHandle, objects::ProjectileWorldObject>,
	ships: HashMap<ColliderHandle, objects::ShipWorldObject>,

	collision_handler: ChannelEventCollector,
	collision_queue: Receiver<CollisionEvent>,
}

// Private methods
impl<'a> World {
	fn remove_projectile(
		&mut self,
		c: ColliderHandle,
	) -> Option<(RigidBody, ProjectileWorldObject)> {
		let p = match self.projectiles.remove(&c) {
			Some(p) => p,
			None => return None,
		};
		let r = self
			.wrapper
			.rigid_body_set
			.remove(
				p.rigid_body,
				&mut self.wrapper.im,
				&mut self.wrapper.collider_set,
				&mut self.wrapper.ij,
				&mut self.wrapper.mj,
				true,
			)
			.unwrap();

		return Some((r, p));
	}

	fn remove_ship(&mut self, h: ShipPhysicsHandle) {
		self.wrapper.rigid_body_set.remove(
			h.0,
			&mut self.wrapper.im,
			&mut self.wrapper.collider_set,
			&mut self.wrapper.ij,
			&mut self.wrapper.mj,
			true,
		);
		self.ships.remove(&h.1);
	}

	/// Add a projectile fired from a ship
	fn add_projectiles(
		&mut self,
		s: ShipPhysicsHandle,
		p: Vec<(object::Projectile, content::GunPoint)>,
	) {
		let mut rng = rand::thread_rng();
		for (projectile, point) in p {
			let r = self.get_ship_body(s).unwrap().1;

			let ship_pos = util::rigidbody_position(r);
			let ship_rot = util::rigidbody_rotation(r);
			let ship_vel = util::rigidbody_velocity(r);
			let ship_ang = ship_rot.angle(Vector2 { x: 0.0, y: 1.0 });

			let pos = ship_pos + (Matrix2::from_angle(-ship_ang) * point.pos.to_vec());

			let spread: Rad<f32> = Deg(
				rng.gen_range(-projectile.content.angle_rng.0..=projectile.content.angle_rng.0)
			)
			.into();

			let vel = ship_vel
				+ (Matrix2::from_angle(-ship_ang + spread)
					* Vector2 {
						x: 0.0,
						y: projectile.content.speed
							+ rng.gen_range(
								-projectile.content.speed_rng..=projectile.content.speed_rng,
							),
					});

			let rigid_body = RigidBodyBuilder::kinematic_velocity_based()
				.translation(vector![pos.x, pos.y])
				.rotation(-ship_ang.0)
				.linvel(vector![vel.x, vel.y])
				.build();

			let collider = match &projectile.content.collider {
				content::ProjectileCollider::Ball(b) => ColliderBuilder::ball(b.radius)
					.sensor(true)
					.active_events(ActiveEvents::COLLISION_EVENTS)
					.build(),
			};

			let rigid_body = self.wrapper.rigid_body_set.insert(rigid_body);
			let collider = self.wrapper.collider_set.insert_with_parent(
				collider,
				rigid_body,
				&mut self.wrapper.rigid_body_set,
			);

			self.projectiles.insert(
				collider.clone(),
				ProjectileWorldObject::new(projectile, rigid_body, collider),
			);
		}
	}

	fn collide_projectile_ship(
		&mut self,
		ct: &content::Content,
		particles: &mut Vec<ParticleBuilder>,
		projectile_h: ColliderHandle,
		ship_h: ColliderHandle,
	) {
		let projectile = self.projectiles.get(&projectile_h);
		let ship = self.ships.get_mut(&ship_h);
		if projectile.is_none() || ship.is_none() {
			return;
		}
		let projectile = projectile.unwrap();
		let ship = ship.unwrap();

		let hit = ship
			.ship
			.handle_projectile_collision(ct, &projectile.projectile);
		let s = ship.physics_handle;
		if hit {
			let pr = self
				.wrapper
				.rigid_body_set
				.get(projectile.rigid_body)
				.unwrap();
			let v = util::rigidbody_velocity(pr).normalize() * projectile.projectile.content.force;
			let pos = util::rigidbody_position(pr);
			let _ = pr;

			let r = self.wrapper.rigid_body_set.get_mut(s.0).unwrap();
			r.apply_impulse_at_point(vector![v.x, v.y], point![pos.x, pos.y], true);

			// Borrow again, we can only have one at a time
			let pr = self
				.wrapper
				.rigid_body_set
				.get(projectile.rigid_body)
				.unwrap();
			let pos = util::rigidbody_position(pr);
			let angle = util::rigidbody_rotation(pr).angle(Vector2 { x: 1.0, y: 0.0 });

			match &projectile.projectile.content.impact_effect {
				None => {}
				Some(x) => {
					let effect = ct.get_effect(*x);
					let (_, sr) = self.get_ship_body(s).unwrap();
					let parent_velocity = util::rigidbody_velocity(pr);
					let target_velocity =
						sr.velocity_at_point(&nalgebra::Point2::new(pos.x, pos.y));

					particles.push(ParticleBuilder::from_content(
						effect,
						pos,
						-angle,
						parent_velocity,
						Vector2 {
							x: target_velocity.x,
							y: target_velocity.y,
						},
					));
				}
			};

			self.remove_projectile(projectile.collider);
		}
	}
}

// Public methods
impl<'a> World {
	/// Create a new physics system
	pub fn new() -> Self {
		let (collision_send, collision_queue) = crossbeam::channel::unbounded();
		let (contact_force_send, _) = crossbeam::channel::unbounded();

		Self {
			wrapper: Wrapper::new(),
			projectiles: HashMap::new(),
			ships: HashMap::new(),
			collision_handler: ChannelEventCollector::new(collision_send, contact_force_send),
			collision_queue,
		}
	}

	/// Add a ship to this physics system
	pub fn add_ship(
		&mut self,
		ct: &content::Content,
		ship: object::Ship,
		position: Point2<f32>,
	) -> ShipPhysicsHandle {
		let ship_content = ct.get_ship(ship.handle);
		let cl = ColliderBuilder::convex_decomposition(
			&ship_content.collision.points[..],
			&ship_content.collision.indices[..],
		)
		// Rotate collider to match sprite
		// (Collider starts pointing east, sprite starts pointing north.)
		.rotation(PI / -2.0)
		.mass(ship_content.mass);
		// TODO: only build colliders once

		let rb = RigidBodyBuilder::dynamic()
			.angular_damping(ship_content.angular_drag)
			.linear_damping(ship_content.linear_drag)
			.translation(vector![position.x, position.y])
			.can_sleep(false);

		let r = self.wrapper.rigid_body_set.insert(rb.build());
		let c = self.wrapper.collider_set.insert_with_parent(
			cl.build(),
			r,
			&mut self.wrapper.rigid_body_set,
		);

		let h = ShipPhysicsHandle(r, c);
		self.ships
			.insert(c, objects::ShipWorldObject::new(ct, ship, h));
		return h;
	}

	/// Step this physics system by `t` seconds
	pub fn step(&mut self, t: f32, ct: &content::Content, particles: &mut Vec<ParticleBuilder>) {
		// Run ship updates
		// TODO: maybe reorganize projectile creation?
		let mut projectiles = Vec::new();
		let mut to_remove = Vec::new();
		for (_, s) in &mut self.ships {
			let r = &mut self.wrapper.rigid_body_set[s.physics_handle.0];
			let c = &mut self.wrapper.collider_set[s.physics_handle.1];

			// TODO: unified step info struct
			s.step(ct, particles, r, c, t);
			if s.controls.guns {
				projectiles.push((s.physics_handle, s.ship.fire_guns()));
			}

			if s.remove_from_world() {
				to_remove.push(s.physics_handle);
				continue;
			}
		}
		for (s, p) in projectiles {
			self.add_projectiles(s, p);
		}
		for s in to_remove {
			self.remove_ship(s);
		}

		// Update physics
		self.wrapper.step(t, &self.collision_handler);

		// Handle collision events
		while let Ok(event) = &self.collision_queue.try_recv() {
			if event.started() {
				let a = event.collider1();
				let b = event.collider2();

				// If projectiles are part of this collision, make sure
				// `a` is one of them.
				let (a, b) = if self.projectiles.contains_key(&b) {
					(b, a)
				} else {
					(a, b)
				};

				let p = self.projectiles.get(&a);
				let s = self.ships.get_mut(&b);
				if p.is_none() || s.is_none() {
					continue;
				}
				self.collide_projectile_ship(ct, particles, a, b);
			}
		}

		// Delete projectiles
		let mut to_remove = Vec::new();
		for (c, p) in &mut self.projectiles {
			p.projectile.tick(t);
			if p.projectile.is_expired() {
				to_remove.push(*c);
			}
		}

		let mut rng = rand::thread_rng();
		for c in to_remove {
			let (pr, p) = self.remove_projectile(c).unwrap();

			match &p.projectile.content.expire_effect {
				None => {}
				Some(x) => {
					let x = ct.get_effect(*x);
					let pos = util::rigidbody_position(&pr);
					let vel = util::rigidbody_velocity(&pr);
					let angle = util::rigidbody_rotation(&pr).angle(Vector2 { x: 1.0, y: 0.0 });

					let velocity = {
						let a = rng
							.gen_range(-x.velocity_scale_parent_rng..=x.velocity_scale_parent_rng);

						let velocity = (x.velocity_scale_parent + a) * vel;

						velocity
					};

					particles.push(ParticleBuilder::from_content(
						x,
						pos,
						-angle,
						velocity,
						Vector2::zero(),
					));
				}
			};
		}
	}

	/// Get a rigid body from a handle
	pub fn get_rigid_body(&self, r: RigidBodyHandle) -> Option<&RigidBody> {
		self.wrapper.rigid_body_set.get(r)
	}

	/// Get a rigid body from a handle
	pub fn get_rigid_body_mut(&mut self, r: RigidBodyHandle) -> Option<&mut RigidBody> {
		self.wrapper.rigid_body_set.get_mut(r)
	}

	/// Get a ship from a handle
	pub fn get_ship_mut(&mut self, s: &ShipPhysicsHandle) -> Option<&mut objects::ShipWorldObject> {
		self.ships.get_mut(&s.1)
	}

	/// Get a ship and its rigidbody from a handle
	pub fn get_ship_body(
		&self,
		s: ShipPhysicsHandle,
	) -> Option<(&objects::ShipWorldObject, &RigidBody)> {
		Some((self.ships.get(&s.1)?, self.wrapper.rigid_body_set.get(s.0)?))
	}

	/// Iterate over all ships in this physics system
	pub fn iter_ship_body(
		&self,
	) -> impl Iterator<Item = (&objects::ShipWorldObject, &RigidBody)> + '_ {
		self.ships.values().map(|x| {
			(
				x,
				self.wrapper.rigid_body_set.get(x.physics_handle.0).unwrap(),
			)
		})
	}

	/// Iterate over all ships in this physics system
	pub fn iter_ships(&'a self) -> impl Iterator<Item = &ShipWorldObject> + '_ {
		self.ships.values()
	}

	/// Iterate over all ships in this physics system
	pub fn iter_projectiles(&'a self) -> impl Iterator<Item = &ProjectileWorldObject> + '_ {
		self.projectiles.values()
	}
}
